import json
import os
import queue
import socket
import sys
import tornado

from threading import Thread
from tornado.ioloop import IOLoop
from tornado import gen
from tornado.websocket import websocket_connect


voc_send_q = queue.Queue()
voc_rcv_q = queue.Queue()

class VocChannel(object):

    def __init__(self):
        token = os.getenv("VOC_COMM_TOKEN")
        host = self.my_ip()
        port = os.getenv("VOC_COMM_PORT")
        proxy = os.getenv("VOC_COMM_PROXY")
        self.url = "wss://" + proxy + "/hostip/" + host + ":" + str(port) + "/voccomm/" + token + "/client?vocmsgs=1"
        self.ws = None
        self.ioloop = IOLoop(make_current=True)
        self.connect()
        self.ioloop.start()

    def my_ip(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('8.8.8.8', 53))
        return s.getsockname()[0]

    @gen.coroutine
    def connect(self):
        print("VOC: Connecting to Vocareum Communication Server...")
        try:
            self.ws = yield websocket_connect(self.url, on_message_callback=self.receiver)
        except Exception:
            print("VOC: Error: Could not connect")
        else:
            print("VOC: Connected")
            self.sender()

    def receiver(self, msg):
        # print("DBG RCVR: MAYBE Adding to Q: {}".format(msg))
        if msg is None:
            print("VOC: ERROR: Vocareum Connection closed")
        try:
            js = json.loads(msg)
            if 'voc' in js:
                if 'severity' in js and js['severity'] == "ERROR":
                    if 'msg' in js and js['msg'] is not None:
                        print("VOC: ERROR: {}".format(js['msg']))
                return
        except: # Exception as e: # json.decoder.JSONDecodeError
            # not vocareum json - should be a real message 
            pass
        # print("DBG RCVR: Adding to Q: {}".format(msg))
        voc_rcv_q.put(msg)

    @gen.coroutine
    def sender(self):
        while True:
            if self.ws is not None:
                # print("VOC DBG: Sender: Wait for CMD...")
                try:
                    msg = voc_send_q.get(block=False)
                    # print("VOC DBG: Sender: Got CMD: {}".format(msg))
                except queue.Empty:
                    yield gen.sleep(0.01)
                else:
                    self.ws.write_message(msg)


class VocCommChannel(object):

    channel = None
    
    def __init__(self, read_timeout=5):
        if VocCommChannel.channel is None:
            VocCommChannel.channel = Thread(target=VocChannel)
            # VocCommChannel.channel.setDaemon(True)
            VocCommChannel.channel.start()
        self.read_timeout = read_timeout

    def read(self, block=True, timeout=None):
        if timeout is None:
            timeout = self.read_timeout
        # print("VOC DBG: Going to READ... ")
        msg = voc_rcv_q.get(block=block, timeout=timeout)
        # print("VOC DBG: Just READ: " + msg)
        return msg        

    def write(self, msg):
        # print("VOC DBG: Going to WRITE: " + msg, flush=True)
        voc_send_q.put(msg)
        # print("VOC DBG: Just WROTE: " + msg, flush=True)

            
    




import time as realtime
import gc

SPEEDUP = 1
tnow = 0


def setup(speedup=1):
    global SPEEDUP
    SPEEDUP = speedup


def time():
    """Returns current clock time."""
    return tnow

def setnow(t):
    global tnow
    tnow = t

def clock(tperiod, tstep=1, tol=0.25):
    """Generator providing time values in sync with real time clock.

    Args:
        tperiod (float): Time interval for clock operation in seconds.
        tstep (float): Time step.
        strict (bool): False turns off real-time synchronization.
        tol (float): Maximum permissible deviation from real time.

    Yields:
        float: The next time step rounded to nearest 10th of a second.

    Raises:
         TCLabClockError: If clock becomes more than `tol` out of phase
             with real time clock.
    """
    global tnow
    setnow(0)
    start_time = SPEEDUP * realtime.time()
    fuzz = 0.003
    k = 0
    while tnow <= tperiod - tstep + tol:
        yield round(tnow, 1)
        k += 1
        if SPEEDUP < 10:
            tsleep = k * tstep\
                     - (SPEEDUP * realtime.time() - start_time)\
                     - SPEEDUP * fuzz
            gcold = gc.isenabled()
            gc.disable()
            try:
                realtime.sleep(max(0, tsleep/SPEEDUP))
            finally:
                if gcold:
                    gc.enable()
            setnow(SPEEDUP * realtime.time() - start_time)
            if abs(tnow - k * tstep) > tol:
                raise RuntimeError("TCLab clock lost synchronization.")
        else:
            tnow = k * tstep

    yield round(tnow, 1)



import datetime
import tornado

from .tclab import TCLab, TCLabModel
from .historian import Historian, Plotter
from .clock import setnow, setup

from ipywidgets import Button, Label, FloatSlider, HBox, VBox, Checkbox, IntText


def actionbutton(description, action, disabled=True):
    """Return button widget with specified label and callback action."""
    button = Button(description=description, disabled=disabled)
    button.on_click(action)

    return button


def labelledvalue(label, value, units=''):
    """Return widget and HBox for label, value, units display."""
    labelwidget = Label(value=label)
    valuewidget = Label(value=str(value))
    unitwidget = Label(value=units)
    box = HBox([labelwidget, valuewidget, unitwidget])

    return valuewidget, box


def slider(label, action, minvalue=0, maxvalue=100, disabled=True):
    """Return slider widget for specified label and action callback."""
    sliderwidget = FloatSlider(description=label, min=minvalue, max=maxvalue)
    sliderwidget.disabled = disabled
    sliderwidget.observe(action, names='value')

    return sliderwidget


class NotebookUI:
    def __init__(self):
        self.timer = tornado.ioloop.PeriodicCallback(self.update, 1000)
        self.lab = None
        self.plotter = None
        self.seconds = 0
        self.firstsession = True

        # Model or real
        self.usemodel = Checkbox(value=False, description='Use model')
        speeduplabel = Label('Speedup')
        self.speedup = IntText(value=1)
        self.speedup.disabled = True
        modelbox = HBox([self.usemodel, speeduplabel, self.speedup])

        # Buttons
        self.connect = actionbutton('Connect', self.action_connect, False)
        self.start = actionbutton('Start', self.action_start)
        self.stop = actionbutton('Stop', self.action_stop)
        self.disconnect = actionbutton('Disconnect', self.action_disconnect)

        buttons = HBox([self.connect, self.start, self.stop, self.disconnect])

        # status
        self.timewidget, timebox = labelledvalue('Timestamp:', 'No data')
        self.sessionwidget, sessionbox = labelledvalue('Session:', 'No data')
        statusbox = HBox([timebox, sessionbox])

        # Sliders for heaters
        self.Q1widget = slider('Q1', self.action_Q1)
        self.Q2widget = slider('Q2', self.action_Q2)

        heaters = VBox([self.Q1widget, self.Q2widget])

        # Temperature display
        self.T1widget, T1box = labelledvalue('T1:', 0, '°C')
        self.T2widget, T2box = labelledvalue('T2:', 0, '°C')

        temperatures = VBox([T1box, T2box])

        self.gui = VBox([modelbox,
                         buttons,
                         statusbox,
                         HBox([heaters, temperatures]),
                         ])

    def update(self):
        """Update GUI display."""
        timestamp = datetime.datetime.now().isoformat(timespec='seconds')
        self.seconds += self.speedup.value
        if self.usemodel.value:
            setnow(self.seconds)
        self.timewidget.value = timestamp
        self.T1widget.value = '{:2.1f}'.format(self.lab.T1)
        self.T2widget.value = '{:2.1f}'.format(self.lab.T2)
        self.plotter.update(self.seconds)

    def action_start(self, widget):
        """Start TCLab operation."""
        self.seconds = 0
        if not self.firstsession:
            self.historian.new_session()
        self.firstsession = False
        self.sessionwidget.value = str(self.historian.session)

        self.start.disabled = True
        self.stop.disabled = False
        self.disconnect.disabled = True

        self.Q1widget.disabled = False
        self.Q2widget.disabled = False

        self.timer.start()

    def action_stop(self, widget):
        """Stop TCLab operation."""
        self.timer.stop()

        self.start.disabled = False
        self.stop.disabled = True
        self.disconnect.disabled = False
        self.Q1widget.disabled = True
        self.Q2widget.disabled = True

    def action_connect(self, widget):
        """Connect to TCLab."""
        if self.usemodel.value:
            self.lab = TCLabModel()
        else:
            self.lab = TCLab()
        self.historian = Historian(self.lab.sources)
        self.plotter = Plotter(self.historian,
                               layout=(('Q1', 'Q2'),
                                       ('T1', 'T2')))
        self.lab.connected = True

        self.usemodel.disabled = True
        self.connect.disabled = True
        self.start.disabled = False
        self.disconnect.disabled = False

    def action_disconnect(self, widget):
        """Disconnect TCLab."""
        self.lab.close()
        self.lab.connected = False

        self.usemodel.disabled = False
        self.connect.disabled = False
        self.disconnect.disabled = True
        self.start.disabled = True

    def action_Q1(self, change):
        """Change heater 1 power."""
        self.lab.Q1(change['new'])

    def action_Q2(self, change):
        """Change heater 2 power."""
        self.lab.Q2(change['new'])
